# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


import unittest
import array

import bl.lib.seq.aligner.io.protobuf_mapping as io
from bl.lib.seq.aligner.mapping import SimpleMapping

class TestProtobufSerialization(unittest.TestCase):

	def protected_compare_seq_maps(self, map_a, map_b):
		self.assertEqual(map_a.get_base_qualities(), map_b.get_base_qualities())
		self.assertEqual(map_a.get_cigar()        , map_b.get_cigar())
		self.assertEqual(map_a.flag         , map_b.flag)
		#self.assertEqual(map_a.isize       , map_b.isize)
		#self.assertEqual(map_a.mpos        , map_b.mpos)
		#self.assertEqual(map_a.mtid        , map_b.mtid)
		self.assertEqual(map_a.get_name()   , map_b.get_name())
		self.assertEqual(map_a.pos          , map_b.pos)
		self.assertEqual(map_a.qual         , map_b.qual)
		self.assertEqual([ t for t in map_a.each_tag() ]          , [ t for t in map_b.each_tag() ])
		self.assertEqual(map_a.tid           , map_b.tid)
		self.assertEqual(map_a.ref_id        , map_b.ref_id)
		self.assertEqual(map_a.get_seq_5(), map_b.get_seq_5())
		self.assertEqual(map_a.get_ascii_base_qual(), map_b.get_ascii_base_qual())
		self.assertEqual(map_a.get_cigar_str(), map_b.get_cigar_str())
		self.assertEqual(map_a.flag_string() , map_b.flag_string())

class TestProtobufSeqMapping(TestProtobufSerialization):

	def setUp(self):
		self.mapping = SimpleMapping()

	def test_simple(self):
		self.__try_and_compare_maps(self.mapping)

	def test_seq(self):
		small_seq = "AGCTNN"
		self.mapping.set_seq_5(small_seq)
		self.mapping.set_base_qualities(array.array("B", [ i + 25 for i in range(len(small_seq)) ]))
		self.__try_and_compare_maps(self.mapping)

	def test_cigar(self):
		self.mapping.set_cigar([ (100, "M") ])
		self.__try_and_compare_maps(self.mapping)
		self.mapping.set_cigar([ (80, "M"), (20, "S") ])
		self.__try_and_compare_maps(self.mapping)

	def test_flag(self):
		self.mapping.flag = 0x04
		self.__try_and_compare_maps(self.mapping)

	def test_name(self):
		self.mapping.set_name("CRESSIA_129:1:1:10029:122606#0/2")
		self.__try_and_compare_maps(self.mapping)

	def test_qual(self):
		self.mapping.qual = 60
		self.__try_and_compare_maps(self.mapping)

	def test_tags(self):
		self.mapping.add_tag( ("XC", 'i', 22) )
		self.mapping.add_tag( ("R2", 'Z', "bla") )
		self.__try_and_compare_maps(self.mapping)

	def test_pos(self):
		self.mapping.pos = 280781134
		self.__try_and_compare_maps(self.mapping)

	def test_tid(self):
		self.mapping.tid = "chr19"
		self.__try_and_compare_maps(self.mapping)

	def test_ref_id(self):
		self.mapping.ref_id = 12
		self.__try_and_compare_maps(self.mapping)
		self.mapping.ref_id = 0
		self.__try_and_compare_maps(self.mapping)

	def test_no_ref_id(self):
		self.mapping.ref_id = None
		self.__try_and_compare_maps(self.mapping)

	def __try_and_compare_maps(self, mapping):
		self.protected_compare_seq_maps(mapping, self.__pipe_map_through(mapping))

	def __pipe_map_through(self, mapping):
		message = io.serialize_seq(mapping)
		return io.unserialize_seq(message)


class TestProtobufPairMapping(TestProtobufSerialization):
	def setUp(self):
		self.pair = (SimpleMapping(), SimpleMapping())
		self.pair[0].set_read1(True)
		self.pair[1].set_read2(True)

		for m in self.pair:
			m.set_paired(True)
			m.set_name("CRESSIA_129:1:1:10605:12550#0")
			m.tid = "chr21"
			m.ref_id = 20
			m.qual = 60
			m.set_cigar( [(100, "M")] )
			m.mtid = "chr21"
			m.m_ref_id = 20

		m1,m2 = self.pair
		m1.pos = m2.mpos = 25839277
		m2.pos = m1.mpos = 25839109
		m1.isize = -268
		m2.isize = 268

	def test_good_pair(self):
		self.__try_and_compare_pair(self.pair)

	def test_isize_zero(self):
		for m in self.pair:
			m.isize = 0
		self.__try_and_compare_pair(self.pair)

	def test_isize_inverted_positions(self):
		# swap positions, invert insert size
		m1, m2 = self.pair
		m2.pos, m1.pos = m1.pos, m2.pos
		m2.mpos, m1.mpos = m1.mpos, m2.mpos
		m1.isize *= -1
		m2.isize *= -1
		self.__try_and_compare_pair(self.pair)

	def test_pair_wo_read1(self):
		pair = (None, self.pair[1])
		m2 = pair[1]
		m2.mtid = None
		m2.m_ref_id = None
		m2.mpos = 0
		m2.isize = 0
		self.__try_and_compare_pair(pair)

	def test_pair_wo_read2(self):
		pair = (self.pair[0], None)
		m1 = pair[0]
		m1.mtid = None
		m1.m_ref_id = None
		m1.mpos = 0
		m1.isize = 0
		self.__try_and_compare_pair(pair)

	def test_both_none(self):
		self.assertRaises(ValueError, io.serialize_pair, (None, None))

	def test_name(self):
		self.pair[0].set_name("my_name/1")
		self.pair[1].set_name("my_name/2")
		# the /1 and /2 will be removed by the serialization
		new_pair = self.__pipe_pair_through(self.pair)
		self.assertEqual("my_name", new_pair[0].get_name())
		self.assertEqual("my_name", new_pair[1].get_name())

	def __try_and_compare_pair(self, pair):
		self.__compare_pair(pair, self.__pipe_pair_through(pair))

	def __pipe_pair_through(self, pair):
		message = io.serialize_pair(pair)
		return io.unserialize_pair(message)

	def __compare_pair(self, pair1, pair2):
		self.assertEqual(len(pair1), len(pair2))
		for i in xrange(len(pair1)):
			if pair1[i] is None:
				self.assertEqual(pair1[i], pair2[i])
			else:
				self.protected_compare_seq_maps(pair1[i], pair2[i])

def suite():
	"""Get a suite with all the tests from this module"""
	suite = unittest.TestSuite()
	suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestProtobufSeqMapping))
	suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestProtobufPairMapping))
	return suite

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import copy

from bl.lib.seq.aligner.mapping import *

class TestMapping(unittest.TestCase):
	def setUp(self):
		self.mapping = SimpleMapping()

	def test_is_paired(self):
		self.assertFalse(self.mapping.is_paired())
		self.mapping.flag |= SAM_FPD
		self.assertTrue(self.mapping.is_paired())

	def test_is_properly_paired(self):
		self.assertFalse(self.mapping.is_properly_paired())
		self.mapping.flag |= SAM_FPP
		self.assertTrue(self.mapping.is_properly_paired())

	def test_is_mapped(self):
		self.assertTrue(self.mapping.is_mapped())
		self.mapping.flag |= SAM_FSU
		self.assertFalse(self.mapping.is_mapped())

	def test_is_unmapped(self):
		self.assertFalse(self.mapping.is_unmapped())
		self.mapping.flag |= SAM_FSU
		self.assertTrue(self.mapping.is_unmapped())

	def test_is_mate_mapped(self):
		self.assertTrue(self.mapping.is_mate_mapped())
		self.mapping.flag |= SAM_FMU
		self.assertFalse(self.mapping.is_mate_mapped())

	def test_is_mate_unmapped(self):
		self.assertFalse(self.mapping.is_mate_unmapped())
		self.mapping.flag |= SAM_FMU
		self.assertTrue(self.mapping.is_mate_unmapped())

	def test_is_on_reverse(self):
		self.assertFalse(self.mapping.is_on_reverse())
		self.mapping.flag |= SAM_FSR
		self.assertTrue(self.mapping.is_on_reverse())

	def test_is_mate_on_reverse(self):
		self.assertFalse(self.mapping.is_mate_on_reverse())
		self.mapping.flag |= SAM_FMR
		self.assertTrue(self.mapping.is_mate_on_reverse())

	def test_is_read1(self):
		self.assertFalse(self.mapping.is_read1())
		self.mapping.flag |= SAM_FR1
		self.assertTrue(self.mapping.is_read1())

	def test_is_read2(self):
		self.assertFalse(self.mapping.is_read2())
		self.mapping.flag |= SAM_FR2
		self.assertTrue(self.mapping.is_read2())

	def test_is_secondary_align(self):
		self.assertFalse(self.mapping.is_secondary_align())
		self.mapping.flag |= SAM_FSC
		self.assertTrue(self.mapping.is_secondary_align())

	def test_set_paired(self):
		self.assertFalse(self.mapping.is_paired())
		self.mapping.set_paired(True)
		self.assertTrue(self.mapping.is_paired())
		self.mapping.set_paired(False)
		self.assertFalse(self.mapping.is_paired())

	def test_set_properly_paired(self):
		self.assertFalse(self.mapping.is_properly_paired())
		self.mapping.set_properly_paired(True)
		self.assertTrue(self.mapping.is_properly_paired())
		self.mapping.set_properly_paired(False)
		self.assertFalse(self.mapping.is_properly_paired())

	def test_set_mapped(self):
		self.assertTrue(self.mapping.is_mapped())
		self.mapping.set_mapped(False)
		self.assertFalse(self.mapping.is_mapped())
		self.mapping.set_mapped(True)
		self.assertTrue(self.mapping.is_mapped())

	def test_set_mate_mapped(self):
		self.assertTrue(self.mapping.is_mate_mapped())
		self.mapping.set_mate_mapped(False)
		self.assertFalse(self.mapping.is_mate_mapped())
		self.mapping.set_mate_mapped(True)
		self.assertTrue(self.mapping.is_mate_mapped())

	def test_set_mate_unmapped(self):
		self.assertFalse(self.mapping.is_paired())
		self.mapping.set_paired(True)
		self.assertTrue(self.mapping.is_paired())
		self.mapping.set_paired(False)
		self.assertFalse(self.mapping.is_paired())

	def test_set_on_reverse(self):
		self.assertFalse(self.mapping.is_on_reverse())
		self.mapping.set_on_reverse(True)
		self.assertTrue(self.mapping.is_on_reverse())
		self.mapping.set_on_reverse(False)
		self.assertFalse(self.mapping.is_on_reverse())

	def test_set_mate_on_reverse(self):
		self.assertFalse(self.mapping.is_mate_on_reverse())
		self.mapping.set_mate_on_reverse(True)
		self.assertTrue(self.mapping.is_mate_on_reverse())
		self.mapping.set_mate_on_reverse(False)
		self.assertFalse(self.mapping.is_mate_on_reverse())

	def test_set_read1(self):
		self.assertFalse(self.mapping.is_read1())
		self.mapping.set_read1(True)
		self.assertTrue(self.mapping.is_read1())
		self.mapping.set_read1(False)
		self.assertFalse(self.mapping.is_read1())

	def test_set_read2(self):
		self.assertFalse(self.mapping.is_read2())
		self.mapping.set_read2(True)
		self.assertTrue(self.mapping.is_read2())
		self.mapping.set_read2(False)
		self.assertFalse(self.mapping.is_read2())

	def test_set_secondary_align(self):
		self.assertFalse(self.mapping.is_secondary_align())
		self.mapping.set_secondary_align(True)
		self.assertTrue(self.mapping.is_secondary_align())
		self.mapping.set_secondary_align(False)
		self.assertFalse(self.mapping.is_secondary_align())

	def test_set_failed_qc(self):
		self.assertFalse(self.mapping.is_failed_qc())
		self.mapping.set_failed_qc(True)
		self.assertTrue(self.mapping.is_failed_qc())
		self.mapping.set_failed_qc(False)
		self.assertFalse(self.mapping.is_failed_qc())

	def test_set_duplicate(self):
		self.assertFalse(self.mapping.is_duplicate())
		self.mapping.set_duplicate(True)
		self.assertTrue(self.mapping.is_duplicate())
		self.mapping.set_duplicate(False)
		self.assertFalse(self.mapping.is_duplicate())

	def test_get_untrimmed_pos_unclipped_forward(self):
		# hit has not been clipped and it's on the forward strand
		self.mapping.pos = 12345
		self.mapping.set_cigar([ (101,'M') ])
		self.assertEqual(self.mapping.pos, self.mapping.get_untrimmed_left_pos())

	def test_get_untrimmed_pos_clipped_forward(self):
		# hit has been clipped to length 90, and it's not on the reverse strand
		self.mapping.pos = 12345
		# read 101 bases long
		self.mapping.set_cigar([ (90,'M'), (11,'S') ])
		self.assertEqual(self.mapping.pos, self.mapping.get_untrimmed_left_pos())
		self.assertEqual(self.mapping.pos + 101-1, self.mapping.get_untrimmed_right_pos())
		self.assertEqual(self.mapping.pos + 90-1, self.mapping.get_right_pos())

	def test_get_untrimmed_pos_unclipped_reverse(self):
		# hit has not been clipped and it's on the reverse strand
		self.mapping.set_on_reverse(True)
		self.mapping.pos = 12345
		self.mapping.set_cigar([ (101,'M') ])
		self.assertEqual(self.mapping.pos, self.mapping.get_untrimmed_left_pos())
		self.assertEqual(self.mapping.pos + 100, self.mapping.get_untrimmed_right_pos())

	def test_get_untrimmed_pos_clipped_reverse(self):
		# hit has been clipped to 35 and it's on the reverse strand
		self.mapping.pos = 12345
		self.mapping.set_seq_5("A"*50) # a 50-base sequence
		self.mapping.set_on_reverse(True)
		self.mapping.set_cigar([ (15,'S'), (35,'M') ])
		self.assertEqual(self.mapping.pos - 15, self.mapping.get_untrimmed_left_pos())
		self.assertEqual(self.mapping.pos + 35-1, self.mapping.get_untrimmed_right_pos())

	def test_get_pos_with_indels(self):
		self.mapping.pos = 15609212
		self.mapping.set_seq_5("A"*101)
		self.mapping.set_cigar([ (15,'S'),(22,'M'),(1,'I'), (63,'M') ])
		self.assertEqual(self.mapping.pos - 15, self.mapping.get_untrimmed_left_pos())
		self.assertEqual(self.mapping.pos + 22+63-1, self.mapping.get_right_pos())
		self.assertEqual(self.mapping.pos + 22+63-1, self.mapping.get_untrimmed_right_pos())

	def test_get_pos_with_indels_and_right_trimming(self):
		self.mapping.pos = 15609212
		self.mapping.set_seq_5("A"*101)
		self.mapping.set_cigar([ (22,'M'),(1,'I'), (63,'M'),(15,'S') ])
		self.assertEqual(self.mapping.pos, self.mapping.get_untrimmed_left_pos())
		self.assertEqual(self.mapping.pos + 22+63-1, self.mapping.get_right_pos())
		self.assertEqual(self.mapping.pos + 22+63+15-1, self.mapping.get_untrimmed_right_pos())

	def test_get_seq_len(self):
		self.mapping.set_seq_5("")
		self.assertEqual(0, self.mapping.get_seq_len())
		self.mapping.set_seq_5("AB" * 10)
		self.assertEqual(20, self.mapping.get_seq_len())

	def test_remove_mate(self):
		m = self.mapping
		m.set_name("p1:read/1")
		m.tid = "chr1"
		m.pos = 12345
		m.set_read1(True)
		m.mtid = "chr1"
		m.mpos = m.pos + 150
		m.set_paired(True)
		m.set_properly_paired(True)
		m.set_mate_mapped(False)
		original = copy.copy(m)

		m.remove_mate()
		self.assertEqual(original.tid, m.tid)
		self.assertEqual(original.pos, m.pos)
		self.assertEqual(None, m.mtid)
		self.assertEqual(0, m.mpos)
		self.assertFalse(m.is_properly_paired())
		self.assertTrue(m.is_mate_mapped()) # make sure the bit ISN'T set

def suite():
	"""Get a suite with all the tests from this module"""
	return unittest.TestLoader().loadTestsFromTestCase(TestMapping)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging
from itertools import izip
import os
import random

from pydoop.pipes import Mapper, InputSplit
from pydoop.utils import jc_configure, jc_configure_int, jc_configure_bool

from bl.lib.seq.aligner.bwa.bwa_aligner import BwaAligner, BWA_INDEX_EXT
import bl.lib.seq.aligner.io.protobuf_mapping as protobuf_mapping
import bl.mr.seq.utils as utils
from bl.mr.lib.hit_processor_chain_link import HitProcessorChainLink
from bl.mr.lib.emit_sam_link import EmitSamLink
from bl.mr.lib.filter_link import FilterLink
from bl.mr.lib.hadoop_event_monitor import HadoopEventMonitor
import bl.lib.tools.deprecation_utils as deprecation_utils
import seqal_app


class MarkDuplicatesEmitter(HitProcessorChainLink):
	def __init__(self, context, event_monitor, next_link = None):
		super(type(self), self).__init__(next_link)
		self.ctx = context
		self.event_monitor = event_monitor

	def __order_pair(self, pair):
		# Order pair such that left-most read is at pos 0.
		# Unmapped reads come after all positions.  None values are last.

		if pair[1] is None:
			ordered_pair = pair
		elif pair[0] is None:
			ordered_pair = (pair[1], pair[0])
		elif pair[1].is_unmapped(): # there are no None values
			ordered_pair = pair
		elif pair[0].is_unmapped():
			ordered_pair = (pair[1], pair[0])
		#else there are no unmapped reads
		elif (pair[0].ref_id, pair[0].get_untrimmed_left_pos(), pair[0].is_on_reverse()) > \
		     (pair[1].ref_id, pair[1].get_untrimmed_left_pos(), pair[1].is_on_reverse()):
			ordered_pair = (pair[1], pair[0])
		else:
			ordered_pair = pair

		return ordered_pair

	def process(self, pair):
		if any(pair):
			# order pair such that left-most read is at pos 0
			ordered_pair = self.__order_pair(pair)

			record = protobuf_mapping.serialize_pair(ordered_pair)
			# emit with the left coord
			key = self.get_hit_key(ordered_pair[0])
			self.ctx.emit(key, record)
			if ordered_pair[0].is_mapped():
				self.event_monitor.count("mapped coordinates", 1)
				# since we ordered the pair, if ordered_pair[0] is unmapped
				# ordered_pair[1] will not be mapped.
				if ordered_pair[1]:
					if ordered_pair[1].is_mapped():
						# a full pair. We emit the coordinate, but with PAIR_STRING as the value
						key = self.get_hit_key(ordered_pair[1])
						self.ctx.emit(key, seqal_app.PAIR_STRING)
						self.event_monitor.count("mapped coordinates", 1)
					else:
						self.event_monitor.count("unmapped reads", 1)
			else:
				self.event_monitor.count("unmapped reads", len(pair))

		# in all cases, forward the original pair to the link in the chain
		if self.next_link:
			self.next_link.process(pair)

	@staticmethod
	def get_hit_key(hit):
		"""Build a key to identify a read hit.
		   To get the proper order in the lexicographic sort, we use a 12-digit
			 field for the position, padding the left with 0s.  12 digits should
			 be enough for any genome :-)

			 We do the same thing for the contig, using its id instead of its name (the tid field).
			 This gives us almost a sorted order in the reducer output; the only entries
			 out of place are the reversed and trimmed reads, since their untrimmed position
			 (used in the key) is different from the reference position.

			 On the other hand, if the read is unmapped, we make a key that starts with the string
			 'unmapped:' and then has a 10-digit random number.  The randomness is inserted so that
			 the unmapped reads may be distributed to the various reducers, instead of having them
			 all send to the same one (since they would all have the same key).
		"""
		if hit.is_mapped():
			if hit.is_on_reverse():
				pos = hit.get_untrimmed_right_pos()
				orientation = 'R'
			else:
				pos = hit.get_untrimmed_left_pos()
				orientation = 'F'
			values = ("%04d" % hit.ref_id, "%012d" % pos, orientation)
		else:
			values = (seqal_app.UNMAPPED_STRING, "%010d" % random.randint(0, 9999999999))
		return ':'.join( values )

class mapper(Mapper):
	"""
	Aligns sequences to a reference genome.

	@input-record: C{key} does not matter (standard LineRecordReader);
	C{value} is a tab-separated text line with 5 fields: ID, read_seq,
	read_qual, mate_seq, mate_qual.

	@output-record: protobuf-serialized mapped pairs (map-reduce job) or alignment
	records in SAM format (map-only job).

	@jobconf-param: C{mapred.reduce.tasks} number of Hadoop reduce tasks to launch.
	If the value of this property is set to 0, then the mapper will directly output
	the mappings in SAM format, like BWA.  If set to a value > 0 the mapper will output
	mappings in the protobuf serialized format for the rmdup reducer.

	@jobconf-param: C{seal.seqal.log.level} logging level,
	specified as a logging module literal.

	@jobconf-param: C{mapred.cache.archives} distributed
	cache entry for the bwa index archive. The entry
	is of the form HDFS_PATH#LINK_NAME. The archive for a given
	chromosome must contain (at the top level, i.e., no directories) all
	files generated by 'bwa index' for that chromosome.

	@jobconf-param: C{seal.seqal.alignment.max.isize}: if the
	inferred isize is greater than this value, Smith-Waterman alignment
	for unmapped reads will be skipped.

	@jobconf-param: C{seal.seqal.pairing.batch.size}: how many
	sequences should be processed at a time by the pairing
	function. Status will be updated at each new batch: therefore,
	lowering this value can help avoid timeouts.

	@jobconf-param: C{seal.seqal.fastq-subformat} Specifies base quality
	score encoding.  Supported types are: 'fastq-sanger' and 'fastq-illumina'.

	@jobconf-param: C{mapred.create.symlink} must be set to 'yes'.

	@jobconf-param: C{seal.seqal.min_hit_quality} mapping quality
	threshold below which the mapping will be discarded.
	"""
	SUPPORTED_FORMATS = "fastq-illumina", "fastq-sanger"
	DEFAULT_FORMAT = "fastq-sanger"
	COUNTER_CLASS = "SEQAL"
	DeprecationMap = {
	  "seal.seqal.log.level":           "bl.seqal.log.level",
	  "seal.seqal.alignment.max.isize": "bl.seqal.alignment.max.isize",
	  "seal.seqal.pairing.batch.size":  "bl.seqal.pairing.batch.size",
	  "seal.seqal.fastq-subformat":     "bl.seqal.fastq-subformat",
	  "seal.seqal.min_hit_quality":     "bl.seqal.min_hit_quality",
	  "seal.seqal.remove_unmapped":     "bl.seqal.remove_unmapped",
	  "seal.seqal.discard_duplicates":  "bl.seqal.discard_duplicates",
	  "seal.seqal.nthreads":            "bl.seqal.nthreads",
	  "seal.seqal.trim.qual":           "bl.seqal.trim.qual",
	}

	def __get_configuration(self, ctx):
		# TODO:  refactor settings common to mapper and reducer
		jc = ctx.getJobConf()

		logger = logging.getLogger("seqal")
		jobconf = deprecation_utils.convert_job_conf(jc, self.DeprecationMap, logger)

		jc_configure(self, jobconf, 'seal.seqal.log.level', 'log_level', 'INFO')
		jc_configure(self, jobconf, "seal.seqal.fastq-subformat", "format", self.DEFAULT_FORMAT)
		jc_configure_int(self, jobconf, 'seal.seqal.alignment.max.isize', 'max_isize', 1000)
		jc_configure_int(self, jobconf, 'seal.seqal.pairing.batch.size', 'batch_size', 10000)
		jc_configure_int(self, jobconf, 'seal.seqal.min_hit_quality', 'min_hit_quality', 0)
		jc_configure_bool(self, jobconf, 'seal.seqal.remove_unmapped', 'remove_unmapped', False)
		jc_configure_int(self, jobconf, 'seal.seqal.nthreads', 'nthreads', 1)
		jc_configure_int(self, jobconf, 'seal.seqal.trim.qual', 'trim_qual', 0)

		try:
			self.log_level = getattr(logging, self.log_level)
		except AttributeError:
			raise ValueError("Unsupported log level: %r" % self.log_level)

		if self.format not in self.SUPPORTED_FORMATS:
			raise_pydoop_exception(
			  "seal.seqal.fastq-subformat must be one of %r" %
			  (self.SUPPORTED_FORMATS,)
			  )

		if self.max_isize <= 0:
			raise ValueError("'seal.seqal.alignment.max.isize' must be > 0, if specified [1000]")

		if self.batch_size <= 0:
			raise ValueError("'seal.seqal.pairing.batch.size' must be > 0, if specified [10000]")

		# minimum qual value required for a hit to be kept.  By default outputs all the
		# hits BWA returns.
		if self.min_hit_quality < 0:
			raise ValueError("'seal.seqal.min_hit_quality' must be >= 0, if specified [0]")

		# number of concurrent threads for main alignment operation
		if self.nthreads <= 0:
			raise ValueError("'seal.seqal.nthreads' must be > 0, if specified [1]")

		# trim quality parameter used by BWA from read trimming.  Equivalent to
		# the -q parameter for bwa align
		if self.trim_qual < 0:
			raise ValueError("'seal.seqal.trim.qual' must be >= 0, if specified [0]")

		if jc.hasKey('mapred.reduce.tasks') and jc.getInt('mapred.reduce.tasks') > 0:
			self.__map_only = False
		else:
			self.__map_only = True

	def __is_last_record(self, k, v):
		return k + len(v) + 2 >= self.split_end


	def get_reference_root(self, ref_dir):
		"""
		Given a directory containing a BWA indexed reference,
		such that all its files have a common name (except the extension),
		this method find the path to the reference including the common name.
		 e.g. my_reference/hg_18.bwt
		      my_reference/hg_18.rsax
		      my_reference/hg_18.sax   => "my_references/hg_18"
		      my_reference/hg_18.pac
		      my_reference/irrelevant_file
		"""
		index_paths = filter(lambda tpl: tpl[1].lstrip('.') in BWA_INDEX_EXT,
		                        map(os.path.splitext, os.listdir(ref_dir)))
		roots = set( zip(*index_paths)[0] )
		if len(roots) == 0:
			raise(ValueError, "Missing references.  Didn't find any files with required extensions (%s) at path %s" % (BWA_INDEX_EXT, ref_dir))
		elif len(roots) != 1:
			raise(ValueError, "multiple references? Found reference roots %s" % (roots,))

		return os.path.join(ref_dir, tuple(roots)[0])

	def __init__(self, ctx):
		super(type(self), self).__init__(ctx)
		self.__get_configuration(ctx)
		logging.basicConfig(level=self.log_level)
		self.event_monitor = HadoopEventMonitor(self.COUNTER_CLASS, logging.getLogger("mapper"), ctx)

		self.aligner = BwaAligner()
		self.aligner.event_monitor = self.event_monitor
		self.aligner.qformat = self.format
		self.aligner.max_isize = self.max_isize
		self.aligner.nthreads = self.nthreads
		self.aligner.trim_qual = self.trim_qual
		self.aligner.mmap_enabled = True

		######## assemble hit processor chain
		chain = FilterLink(self.event_monitor)
		chain.remove_unmapped = self.remove_unmapped
		chain.min_hit_quality = self.min_hit_quality
		if self.__map_only:
			chain.set_next( EmitSamLink(ctx, self.event_monitor) )
		else:
			chain.set_next( MarkDuplicatesEmitter(ctx, self.event_monitor) )
		self.aligner.hit_visitor = chain

		######## set the path to the reference index
		self.ref_archive = utils.get_ref_archive(ctx.getJobConf())
		self.aligner.reference = self.get_reference_root(self.ref_archive)

		# part of the code is a workaround for accumulating records, see #331
		isplit = InputSplit(ctx.getInputSplit())
		self.split_end = isplit.offset + isplit.length

	def map(self, ctx):
		# Accumulates reads in self.pairs, until batch size is reached or
		# until the input is finished.	At that point it calls run_alignment
		# and emits the output.
		k = struct.unpack(">q", ctx.getInputKey())[0]
		v = ctx.getInputValue()
		self.aligner.load_pair_record(v.split("\t"))
		is_last_record = self.__is_last_record(k, v)
		if self.aligner.get_batch_size() >= self.batch_size or is_last_record:
			self.aligner.run_alignment()
			self.aligner.clear_batch()

		if is_last_record:
			self.aligner.release_resources()



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

class EventMonitor(object):
	"""
	Interface to count events and measure time spent through counters.
	"""
	def __init__(self): pass

	def start(self, event_name):
		"""
		Start timing an event.	Stops timing at the corresponding call to
		"stop" with the same event_name
		"""
		abstract

	def stop(self, event_name):
		"""
		Stop the timer for event_name started with "start"
		"""
		abstract

	class TimingBlock(object):
		def __init__(self, monitor, event_name):
			self.__monitor = monitor
			self.__event_name = event_name

		def __enter__(self):
			self.__monitor.start(self.__event_name)
			return self.__monitor

		def __exit__(self, exception_type, exception_val, exception_tb):
			self.__monitor.stop(self.__event_name)
			return False

	def time_block(self, event_name):
		"""
		Use in a with statement to time a block of code.  Example:
		  with monitor.time_block("my_event"):
			  do_something()
		"""
		return type(self).TimingBlock(self, event_name)

	def stop_batch(self, event_name, offset, n):
		abstract

	def count(self, event_name, value=1):
		"""
		Increment counter "event_name" by "value" (default 1).
		The counter will be created automatically if it
		doesn't already exist.
		"""
		abstract

	def has_counter(self, event_name):
		"""
		Check whether a counter for "event_name" has already been created.
		"""
		abstract

	def add_counter(self, event_name, display_name=None):
		"""
		Explicitly create a counter for "event_name", optionally
		setting a display name for reporting.	If no display_name is
		provided, then event_name will be used instead.

		Since counters can be created on-demand by "count", you usually
		won't need to use this method unless you want to set the display_name.
		"""
		abstract

	#################################
	# logging methods
	#################################
	def new_status(self, status_msg):
		"""
		Report that the job has entered a new state.
		"""
		abstract

	def log_debug(self, *args):
		"""
		Log a message if log level is at debug or higher.
		"""
		abstract

	def log_info(self, *args):
		"""
		Log a message if log level is at info or higher.
		"""
		abstract

	def log_warning(self, *args):
		"""
		Log a message if log level is at warning or higher.
		"""
		abstract

	def log_error(self, *args):
		"""
		Log a message if log level is at error or higher.
		"""
		abstract

	def log_critical(self, *args):
		"""
		Log a message if log level is at critical or higher.
		"""
		abstract


class QuietMonitor(EventMonitor):
	"""
	Implementation of EventMonitor that doesn't do anything.
	"""
	def start(self, event_name): pass
	def stop(self, event_name): pass
	def stop_batch(self, event_name, offset, n): pass
	def count(self, event_name, value=1): pass
	def has_counter(self, event_name): pass
	def add_counter(self, event_name, display_name=None): pass
	def new_status(self, status_msg): pass
	def log_debug(self, *args): pass
	def log_info(self, *args): pass
	def log_warning(self, *args): pass
	def log_error(self, *args): pass
	def log_critical(self, *args): pass



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import re

from pydoop.utils import make_input_split


# import pydoop innards
try:
	# version < 0.5
	from pydoop._pipes import get_JobConf_object
except ImportError:
	# version >= 0.5
	import pydoop
	pp = pydoop.import_version_specific_module("_pipes")
	get_JobConf_object = pp.get_JobConf_object


from bl.lib.seq.aligner.sam_mapping import SAMMapping
from bl.mr.test_utils import reduce_context
from bl.mr.seq.seqal.reducer import reducer
from bl.mr.seq.seqal.seqal_app import PAIR_STRING
import bl.lib.seq.aligner.io.protobuf_mapping as proto
import bl.lib.seq.aligner.sam_flags as sam_flags
import test_utils # specific to seqal

class TestSeqalReducer(unittest.TestCase):

	def setUp(self):
		self.__jc = get_JobConf_object({})
		self.__ctx = reduce_context(self.__jc, [])
		self.__reducer = reducer(self.__ctx)
		self.__reducer.discard_duplicates = True
		self.__clean_reducer = reducer(self.__ctx) # unmodified

	def test_emit_on_left_key(self):
		# load pair 1
		p = test_utils.pair1()
		# use the first read to create the map-reduce key
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__reducer.reduce(self.__ctx)
		self.__ensure_only_pair1_emitted()

	def test_no_emit_on_right_key(self):
		# load pair 1
		p = test_utils.pair1()
		# use the SECOND read to create the map-reduce key
		self.__ctx.add_value(test_utils.make_key(p[1]), PAIR_STRING)
		self.__reducer.reduce(self.__ctx)
		# we should have no output
		self.assertEqual(0, len(self.__ctx.emitted.keys()))

	def test_duplicate_pairs(self):
		# Two identical pairs.  Ensure only one is emitted
		p = test_utils.pair1()
		# use the first read to create the map-reduce key
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p)) # add it twice
		self.__reducer.reduce(self.__ctx)
		self.assertEqual(1, len(self.__ctx.emitted.keys()))
		self.assertEqual(2, len(self.__ctx.emitted.values()[0])) # two SAM records associated with the same key
		self.__ensure_only_pair1_emitted()
		# check counter
		if self.__ctx.counters.has_key(self.__frag_counter_name()):
			self.assertEqual(0, self.__ctx.counters[self.__frag_counter_name()])
		self.assertTrue(self.__ctx.counters.has_key(self.__pair_counter_name()))
		self.assertEqual(1, self.__ctx.counters[self.__pair_counter_name()])

	def test_duplicate_pairs_no_discard(self):
		# Two identical pairs.  Ensure only one is emitted
		p = test_utils.pair1()
		# use the first read to create the map-reduce key
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p)) # add it twice
		self.__reducer.discard_duplicates = False
		self.__reducer.reduce(self.__ctx)
		self.assertEqual(1, len(self.__ctx.emitted.keys()))
		self.assertEqual(4, len(self.__ctx.emitted.values()[0])) # four SAM records associated with the same key
		flags = map(lambda sam: int(*re.match("(\d+).*", sam).groups(1)), self.__ctx.emitted.values()[0])
		# ensure we have two marked as duplicates
		self.assertEqual(2, len(filter(lambda flag: flag & sam_flags.SAM_FDP, flags)) )
		# ensure we have two NOT marked as duplicates
		self.assertEqual(2, len(filter(lambda flag: flag & sam_flags.SAM_FDP == 0, flags)) )
		# check counter
		if self.__ctx.counters.has_key(self.__frag_counter_name()):
			self.assertEqual(0, self.__ctx.counters[self.__frag_counter_name()])
		self.assertTrue(self.__ctx.counters.has_key(self.__pair_counter_name()))
		self.assertEqual(1, self.__ctx.counters[self.__pair_counter_name()])


	def test_duplicate_pairs_right_key(self):
		# Two identical pairs on the right key
		# Ensure nothing is emitted
		p = test_utils.pair1()
		# use the first read to create the map-reduce key
		self.__ctx.add_value(test_utils.make_key(p[1]), PAIR_STRING)
		self.__ctx.add_value(test_utils.make_key(p[1]), PAIR_STRING) # add it twice
		self.__reducer.reduce(self.__ctx)
		self.assertEqual(0, len(self.__ctx.emitted.keys()))
		# check counter
		if self.__ctx.counters.has_key(self.__pair_counter_name()):
			self.assertEqual(0, self.__ctx.counters[self.__pair_counter_name()])
		if self.__ctx.counters.has_key(self.__frag_counter_name()):
			self.assertEqual(0, self.__ctx.counters[self.__frag_counter_name()])

	def test_duplicate_fragments_read1(self):
		# load pair 1
		p = list(test_utils.pair1())
		p = test_utils.erase_read2(p)
		p0 = p[0]
		# insert the pair into the context, twice
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__reducer.reduce(self.__ctx)
		self.assertEqual(1, len(self.__ctx.emitted.keys()))
		self.assertEqual(1, len(self.__ctx.emitted.values()[0])) # only one SAM record associated with the key
		short_name = p0.get_name()[0:-2]
		self.assertEqual(short_name, self.__ctx.emitted.keys()[0])
		self.assertTrue( re.match("\d+\s+%s\s+%d\s+.*" % (p0.tid, p0.pos), self.__ctx.emitted[short_name][0]) )
		# check counter
		self.assertFalse(self.__ctx.counters.has_key(self.__pair_counter_name()))
		self.assertTrue(self.__ctx.counters.has_key(self.__frag_counter_name()))
		self.assertEqual(1, self.__ctx.counters[self.__frag_counter_name()])

	def test_duplicate_fragments_read1_no_discard(self):
		# load pair 1 and erase its second read
		p = list(test_utils.pair1())
		p = test_utils.erase_read2(p)
		p0 = p[0]
		# insert the pair into the context, twice
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__reducer.discard_duplicates = False
		self.__reducer.reduce(self.__ctx)
		self.assertEqual(1, len(self.__ctx.emitted.keys()))
		self.assertEqual(2, len(self.__ctx.emitted.values()[0])) # Two SAM records associated with the key
		short_name = p0.get_name()[0:-2]
		self.assertEqual(short_name, self.__ctx.emitted.keys()[0])
		flags = map(lambda sam: int(*re.match("(\d+).*", sam).groups(1)), self.__ctx.emitted.values()[0])
		# ensure we have one marked as duplicate
		self.assertEqual(1, len(filter(lambda flag: flag & sam_flags.SAM_FDP, flags)) )
		# and ensure we have one NOT marked as duplicates
		self.assertEqual(1, len(filter(lambda flag: flag & sam_flags.SAM_FDP == 0, flags)) )

		# check counter
		self.assertFalse(self.__ctx.counters.has_key(self.__pair_counter_name()))
		self.assertTrue(self.__ctx.counters.has_key(self.__frag_counter_name()))
		self.assertEqual(1, self.__ctx.counters[self.__frag_counter_name()])

	def test_empty_read1(self):
		# Ensure the reducer raises an exception if the pair[0] is None
		p = test_utils.erase_read1(list(test_utils.pair1()))
		self.__ctx.add_value(test_utils.make_key(p[1]), proto.serialize_pair(p))
		self.assertRaises(ValueError, self.__reducer.reduce, self.__ctx)

	def test_fragment_with_duplicate_in_pair_1(self):
		# Ensure the reducer catches a fragment duplicate of pair[0]
		p = list(test_utils.pair1())
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		test_utils.erase_read2(p)
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__reducer.reduce(self.__ctx)
		# now ensure that the pair was emitted, but not the fragment
		self.__ensure_only_pair1_emitted()
		self.assertEqual(1, len(self.__ctx.emitted.keys()))
		self.assertEqual(2, len(self.__ctx.emitted.values()[0])) # two SAM records associated with the key (for the pair)
		# check counter
		self.assertFalse(self.__ctx.counters.has_key(self.__pair_counter_name()))
		self.assertTrue(self.__ctx.counters.has_key(self.__frag_counter_name()))
		self.assertEqual(1, self.__ctx.counters[self.__frag_counter_name()])

	def test_fragment_with_duplicate_in_pair_2(self):
		# Ensure the reducer catches a fragment duplicate of pair[1].
		p = list(test_utils.pair1())
		# Insert the pair into the context
		self.__ctx.add_value(test_utils.make_key(p[1]), PAIR_STRING)
		# Remove the first read from the pair, reorder so that the None is at index 1,
		# the serialize and insert into the context.
		test_utils.erase_read1(p)
		self.__ctx.add_value(test_utils.make_key(p[1]), proto.serialize_pair( (p[1], None) ))
		self.__reducer.reduce(self.__ctx)
		# now ensure that nothing was emitted.  The pair isn't emitted because
		# the key refers to read2, and the fragment isn't emitted because it's a duplicate of
		# the one in the pair.
		self.assertEqual(0, len(self.__ctx.emitted.keys()))
		# check counter
		self.assertFalse(self.__ctx.counters.has_key(self.__pair_counter_name()))
		self.assertTrue(self.__ctx.counters.has_key(self.__frag_counter_name()))
		self.assertEqual(1, self.__ctx.counters[self.__frag_counter_name()])

	def test_fragment_with_duplicate_in_pair_1_no_discard(self):
		# Ensure the reducer catches a fragment duplicate of pair[0]
		p = list(test_utils.pair1())
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		p = test_utils.erase_read2(p)
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__reducer.discard_duplicates = False
		self.__reducer.reduce(self.__ctx)
		# now ensure that both were emitted, but the fragment is marked as duplicate
		self.__ensure_pair1_emitted()
		self.assertEqual(1, len(self.__ctx.emitted.keys()))
		self.assertEqual(3, len(self.__ctx.emitted.values()[0])) # 3 SAM records associated with the key (for the pair)

		# make sure we have a read with the duplicate flag set
		regexp = "(\d+)\s+.*"
		flags = [ int(re.match(regexp, value).group(1)) for value in self.__ctx.emitted.values()[0] ]
		dup_flags = [ flag for flag in flags if flag & sam_flags.SAM_FDP ]
		self.assertEqual(1, len(dup_flags))
		f = dup_flags[0]
		self.assertTrue( f & sam_flags.SAM_FR1 > 0 ) # ensure the duplicate read is r1
		self.assertTrue( f & sam_flags.SAM_FPD == 0 ) # ensure the duplicate read is unpaired

		# check counter
		self.assertFalse(self.__ctx.counters.has_key(self.__pair_counter_name()))
		self.assertTrue(self.__ctx.counters.has_key(self.__frag_counter_name()))
		self.assertEqual(1, self.__ctx.counters[self.__frag_counter_name()])


	def test_default_discard_duplicates_setting(self):
		self.assertFalse(self.__clean_reducer.discard_duplicates)

	def test_unmapped2(self):
		p = test_utils.pair1()
		p[1].set_mapped(False)
		p[0].set_mate_mapped(False)
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__reducer.reduce(self.__ctx)
		self.assertEqual(2, len(self.__ctx.emitted.values()[0]))

	def test_unmapped1(self):
		p = test_utils.pair1()
		p[0].set_mapped(False)
		p[1].set_mate_mapped(False)
		# Having an unmapped read before a mapped read is not allowed.  This should
		# raise an exception
		# The key is meaningless
		self.__ctx.add_value(test_utils.make_key(p[1]), proto.serialize_pair(p))
		self.assertRaises(ValueError, self.__reducer.reduce, self.__ctx)

	def test_unmapped_pair(self):
		p = test_utils.pair1()
		for i in 0,1:
			p[i].set_mapped(False)
			p[i].set_mate_mapped(False)
		self.__ctx.add_value(test_utils.make_key(p[0]), proto.serialize_pair(p))
		self.__reducer.reduce(self.__ctx)
		self.assertEqual(1, len(self.__ctx.emitted.keys()))
		self.assertEqual(2, len(self.__ctx.emitted.values()[0]))

	def test_rmdup_bug(self):
		test_case_data = [
"HWI-ST200R_251:5:1208:19924:124635#GCCAAT\t83\t20\t6181935\t60\t5S96M\t=\t6181919\t-112\tAAGTGGAAGATTTGGGAATCTGAGTGGATTTGGTAACAGTAGAGGGGTGGATCTGGCTTGGAAAACAATCGAGGTACCAATATAGGTGGTAGATGAATTTT\t?<?AADADBFBF<EHIGHGGGEAF3AF<CHGGDG9?GHFFACDHH)?@AHEHHIIIIE>A=A:?);B27@;@?>,;;C(5::>>>@5:()4>@@@######\tXC:i:96\tXT:A:U\tNM:i:1\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:1\tXO:i:0\tXG:i:0\tMD:Z:13G82\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:######@@@>4)(:5@>>>::5(C;;,>?@;@72B;)?:A=A>EIIIIHHEHA@?)HHDCAFFHG?9GDGGHC<FA3FAEGGGHGIHE<FBFBDADAA?<?",
"HWI-ST200R_251:5:1208:19924:124635#GCCAAT\t163\t20\t6181919\t60\t101M\t=\t6181935\t112\tCTGAGCACACCAAAATTCATCTACCACCTATATTGGTACCTCGATTGTTTTCCAAGCCAGATCCACACCTCTACTGTTACCAAATCCACTCAGATTCCCAA\t@@@FFFDDFHG??;EEH>HHGIGHEGCGEGGIGJG31?DDBBD>FGG@HG??DFBBADFAGII3@EH;;CEHECBB7?>CE.;...5>ACDDA:C:;>:>?\tXT:A:U\tNM:i:2\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:2\tXO:i:0\tXG:i:0\tMD:Z:29G36C34\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:@@@FFFDDFHG??;EEH>HHGIGHEGCGEGGIGJG31?DDBBD>FGG@HG??DFBBADFAGII3@EH;;CEHECBB7?>CE.;...5>ACDDA:C:;>:>?",
"HWI-ST200R_251:6:2207:18561:163438#GCCAAT\t83\t20\t6181938\t60\t8S93M\t=\t6181919\t-112\tAAAATTCATCTACCACCTATATTGGTACCTCGATTGTTTTCCAAGCCAGATCCACCCCTCTACTGTTACCAAATCCACTCAGATTCCCAAATCTTCCACTT\t@@@DDFDFHHHHHJJJEHGGHIHHAEGHJJIJJFGGHGIDIGIJJ?BBGGGIIIJJIJGFHGIJEC(=3?C;?B9?@C>CECECAA(;;3>C#########\tXC:i:93\tXT:A:U\tNM:i:2\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:2\tXO:i:0\tXG:i:0\tMD:Z:10G36C45\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:#########C>3;;(AACECEC>C@?9B?;C?3=(CEJIGHFGJIJJIIIGGGBB?JJIGIDIGHGGFJJIJJHGEAHHIHGGHEJJJHHHHHFDFDD@@@",
"HWI-ST200R_251:6:2207:18561:163438#GCCAAT\t163\t20\t6181919\t60\t101M\t=\t6181938\t112\tCTGAGCACACCAAAATTCATCTACCACCTATATTGGTACCTCGATTGTTTTCCAAGCCAGATCCACACCTCTACTGTTACCAAATCCACTCAGATTCCCAA\t@CCFFDDFHHHHHIJJJIIJJJIJJIIJGJIIIJII?DGHIGHDGHIIIJIJIJIIDCHGIJIIGGHIFEHHHHFFFFFDC.6.66;@CCCDCCDC>CCCA\tXT:A:U\tNM:i:2\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:2\tXO:i:0\tXG:i:0\tMD:Z:29G36C34\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:@CCFFDDFHHHHHIJJJIIJJJIJJIIJGJIIIJII?DGHIGHDGHIIIJIJIJIIDCHGIJIIGGHIFEHHHHFFFFFDC.6.66;@CCCDCCDC>CCCA",
		]
		self.__reducer.discard_duplicates = False
		sams = map(SAMMapping, test_case_data)
		left_key = "0020:000006181919:F"
		pairs = ( (sams[1], sams[0]), (sams[3], sams[2]) )

		# add the pairs to the context
		self.__ctx.add_value(left_key, proto.serialize_pair(pairs[0]))
		self.__ctx.add_value(left_key, proto.serialize_pair(pairs[1]))
		self.__reducer.reduce(self.__ctx)

		self.assertEqual(2, len(self.__ctx.emitted))
		self.assertEqual([2,2], map(len, self.__ctx.emitted.itervalues()))
		key = "HWI-ST200R_251:6:2207:18561:163438#GCCAAT"
		good_pair_sam = self.__ctx.emitted[key]
		for read in good_pair_sam:
			mapping = SAMMapping("\t".join( (key,read) ))
			self.assertFalse(mapping.is_duplicate())

		key = "HWI-ST200R_251:5:1208:19924:124635#GCCAAT"
		dup_pair_sam = self.__ctx.emitted[key]
		for read in dup_pair_sam:
			mapping = SAMMapping("\t".join( (key,read) ))
			self.assertTrue(mapping.is_duplicate())

	def test_fw_rev_missed_dup_pair(self):
		"""
		Here we have two duplicate pairs, as detected by Picard.
		The reads 10364 have flags 'pPR1', 'pPr2'.
		The reads 138222 have flags 'pPR2d', 'pPr1d'.
		10364/1 has no trimming and is on the fw strand.  It should have a key position 88888399:F.
		10364/2 is also on the rev strand.  We expect a key position of 88888421+53 = 88888474:R.
		138222/1 is on the rev strand.  We expect a key position of 88888404+70 = 88888474:R.
		138222/2 has been trimmed but is on the fw strand. It should have a key position 88888399:F.
		"""
		test_case_data = [
			"HWI-ST332_97:3:66:16214:10364#0\t99\t10\t88888399\t46\t76M\t=\t88888421\t76\tTGATTTTGCTCCATTGTCTTCTAGCTTGTGTTATGCCTGTTGAAAGTACAAAATCATTCTGGAAGCTTATCTATTG\tHGHHHHHHFHEHHHHGHHFHHHGGHHHHHHHHHHEHDHGEHFHHHHHHHGGHHHHHHHHHHBHBBEGG=GFFFF@F",
			"HWI-ST332_97:3:66:16214:10364#0\t147\t10\t88888421\t46\t22S54M\t=\t88888399\t-76\tTGATTCCGCTCCATGTGCCTCGAGCTTGTGTTATGCCTGTTGAAAGTACAAAATCATTCTGGAAGCTTATCTATTG\t#######################AA@EHGHEHHHHHHHHHHHHHHHEHHHHHHFHFHHHHHHHHHHHHHHHGHHFH",
			"HWI-ST332_97:3:7:10556:138222#0\t1107\t10\t88888404\t23\t5S71M\t=\t88888399\t-76\tTGATGTTGCTCCATTGTCTTCTAGCTTGTGTTATGCCTGTTGAAAGTACAAAATCATTCTGGAAGCTTATCTATTG\t######C<BCC:B,DDDC=BD3@CB8B?DBD@E@EEEEECED@CDB=8C7A@D=DEDEDCDBECDE<>;;,17,45",
			"HWI-ST332_97:3:7:10556:138222#0\t1187\t10\t88888399\t15\t50M26S\t=\t88888404\t76\tTGATTTTGCTCCATTGTCTTCTAGCTTGTGTTATGCCTGTTGAAAGTACAAAATCCGTCTGGTTGCTTCTATTTTG\tCC7EEFFF@FHHFHHG?GGF:>4.7GD8DC@D>CCFGG@GGG5GG4<CB###########################",
		]
		sams = map(SAMMapping, test_case_data)
		pair1 = sams[0:2]
		pair2 = sams[2:]

		left_key = "0010:000088888399:F"
		self.__ctx.add_value(left_key, proto.serialize_pair(pair1))
		self.__ctx.add_value(left_key, proto.serialize_pair(pair2))
		self.__reducer.reduce(self.__ctx)

		# verify emitted data.  2 keys (one per read name) and 2 reads for each key
		self.assertEqual(2, len(self.__ctx.emitted))
		self.assertEqual([2,2], map(len, self.__ctx.emitted.itervalues()))
		key = "HWI-ST332_97:3:66:16214:10364#0"
		good_pair_sam = self.__ctx.emitted[key]
		for read in good_pair_sam:
			mapping = SAMMapping("\t".join( (key,read) ))
			self.assertFalse(mapping.is_duplicate())
		key = "HWI-ST332_97:3:7:10556:138222#0"
		dup_pair_sam = self.__ctx.emitted[key]
		for read in dup_pair_sam:
			mapping = SAMMapping("\t".join( (key,read) ))
			self.assertTrue(mapping.is_duplicate())

		self.setUp() # clean-up and repeat for the right key
		right_key = "0010:000088888474:R"
		self.__ctx.add_value(right_key, PAIR_STRING)
		self.__ctx.add_value(right_key, PAIR_STRING)
		self.__reducer.reduce(self.__ctx)
		# verify no data emitted
		self.assertEqual(0, len(self.__ctx.emitted))


	def test_rmdup_clipped_unpaired(self):
		test_case_data = [
"HWI-ST200R_251:5:1208:19924:124635#GCCAAT\t82\t20\t6181935\t60\t5S96M\t*\t0\t0\tAAGTGGAAGATTTGGGAATCTGAGTGGATTTGGTAACAGTAGAGGGGTGGATCTGGCTTGGAAAACAATCGAGGTACCAATATAGGTGGTAGATGAATTTT\t?<?AADADBFBF<EHIGHGGGEAF3AF<CHGGDG9?GHFFACDHH)?@AHEHHIIIIE>A=A:?);B27@;@?>,;;C(5::>>>@5:()4>@@@######\tXC:i:96",
"HWI-ST200R_251:6:2207:18561:163438#GCCAAT\t82\t20\t6181930\t60\t101M\t*\t0\t0\tAAGTGGAAGATTTGGGAATCTGAGTGGATTTGGTAACAGTAGAGGGGTGGATCTGGCTTGGAAAACAATCGAGGTACCAATATAGGTGGTAGATGAATTTT\t@CCFFDDFHHHHHIJJJIIJJJIJJIIJGJIIIJII?DGHIGHDGHIIIJIJIJIIDCHGIJIIGGHIFEHHHHFFFFFDC.6.66;@CCCDCCDC>CCCA",
		]
		self.__reducer.discard_duplicates = False
		sams = map(SAMMapping, test_case_data)
		left_key = "0020:000006181930:R"
		pairs = ( (sams[0], None), (sams[1], None) )

		# add the pairs to the context
		self.__ctx.add_value(left_key, proto.serialize_pair(pairs[0]))
		self.__ctx.add_value(left_key, proto.serialize_pair(pairs[1]))
		self.__reducer.reduce(self.__ctx)

		self.assertEqual(2, len(self.__ctx.emitted))
		self.assertEqual([1,1], map(len, self.__ctx.emitted.itervalues()))

		key = "HWI-ST200R_251:6:2207:18561:163438#GCCAAT"
		good_read_sam = self.__ctx.emitted[key][0]
		mapping = SAMMapping("\t".join( (key,good_read_sam) ))
		self.assertFalse(mapping.is_duplicate())	

		key = "HWI-ST200R_251:5:1208:19924:124635#GCCAAT"
		dup_read_sam = self.__ctx.emitted[key][0]
		mapping = SAMMapping("\t".join( (key,dup_read_sam) ))
		self.assertTrue(mapping.is_duplicate())	


	def __ensure_pair1_emitted(self):
		p = test_utils.pair1()
		# Now we expect a SAM entry for each of the two pairs.
		# At the moment, the SAM formatter emits the read name as the key, and the
		# rest of the SAM record as the value.  Remember that the protobuff serialization
		# removes the read number ("/1" or "/2") from the read name.
		short_name = p[0].get_name()[0:-2] # mapping name without the read number
		self.assertTrue( self.__ctx.emitted.has_key(short_name)  )
		self.assertTrue(len(self.__ctx.emitted[short_name]) >= 2 ) # at least two SAM records emitted


	def __ensure_only_pair1_emitted(self):
		self.__ensure_pair1_emitted()
		p = test_utils.pair1()
		short_name = p[0].get_name()[0:-2] # mapping name without the read number
		self.assertEqual( [short_name], self.__ctx.emitted.keys() )
		self.assertEqual(2, len(self.__ctx.emitted[short_name])) # two SAM records emitted
		regexp = "\d+\s+%s\s+(\d+)\s+.*" % p[0].tid # all reads have the same tid.  Match the position
		emitted_positions = map(lambda sam: int(*re.match(regexp, sam).groups(1)), self.__ctx.emitted[short_name])
		self.assertEqual( [p[0].pos, p[1].pos], sorted(emitted_positions) ) # ensure we have both positions
		emitted_flags = map(lambda sam: int(*re.match("(\d+).*", sam).groups(1)), self.__ctx.emitted[short_name])
		# ensure all the reads we found are flagged as mapped
		self.assertTrue( all(map(lambda flag: flag & (sam_flags.SAM_FSU | sam_flags.SAM_FMU) == 0, emitted_flags)) )

	def __pair_counter_name(self):
		return ':'.join( (self.__reducer.COUNTER_CLASS, "DUPLICATE PAIRS") )

	def __frag_counter_name(self):
		return ':'.join( (self.__reducer.COUNTER_CLASS, "DUPLICATE FRAGMENTS") )


def suite():
	return unittest.TestLoader().loadTestsFromTestCase(TestSeqalReducer)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())
